/////////////////////////////////////////////////////////////////////////////////////
///
/// algebraicexpr.cpp
///
/// Acme Software Works, Inc.
///
/// Created January 27, 2004 by Don Dugger
///
/// <PRE>
/// Copyright 2019 Acme Software Works, Inc.
/// 
/// Redistribution and use in source and binary forms, with or without
/// modification, are permitted provided that the following conditions are met:
/// 
/// 1. Redistributions of source code must retain the above copyright notice,
///    this list of conditions and the following disclaimer.
/// 
/// 2. Redistributions in binary form must reproduce the above copyright notice,
///    this list of conditions and the following disclaimer in the documentation
///    and/or other materials provided with the distribution.
/// 
/// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
/// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
/// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
/// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
/// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
/// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
/// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
/// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
/// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
/// OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
/// </PRE>
/////////////////////////////////////////////////////////////////////////////////////

#include "algebraicexpr.h"

///////////////////////////////////////////////////////////////

int AlgebraicExpr::token_prec[] = {
#include "token_prec.cpp"
};

///////////////////////////////////////////////////////////////

AlgebraicExpr::TOKEN_CATEGORY AlgebraicExpr::token_cat[] = {
#include "token_cat.cpp"
};

///////////////////////////////////////////////////////////////

void AlgebraicExpr::printTok(int token)
{
#include "token_print.cpp"
} // End of printTok()

///////////////////////////////////////////////////////////////

double AlgebraicExpr::solve(string ex)
{
    TOKEN token = (TOKEN)0;
    int parens = 0;
    bool done = false;
    TOKEN_TYPE last_oper = OPERATOR;
    ExprElement* oper;
    vector<ExprElement*> opers;
    vector<ExprElement*> post;
    string token_str;
    // Used to indicate unassigned variables
    bool must_be_asignment = false;
    // Used to tell the name of the non-existent
    // variable in the error message
    string token_str_added;

    // If the user did not give us a 
    // variable list then we need to
    // allocate one.
    if(var_list == NULL) {
        var_list = new map<string,double>;
        var_list_allocated = true;
    }
    this->expr = ex;
    cur = 0;
    // Remove spaces
    for (int i=0;i<expr.size();++i) {
        if ( expr[i] == ' ' ) expr.erase(i,1);
    }
    if (debug_flag) cout << "PARSING> " << expr << '\n';
    do {
        if (debug_flag == 2 && token != 0 ) {
            cout << endl << "--------post->token-----------" << endl;
            vector<ExprElement*>::iterator pit = post.begin();
            while(pit != post.end() ) {
                ExprElement* tmp = *pit;
                cout << tmp->value << " ";
                printTok(tmp->token);
                ++pit;
            }
            cout << "--------opers->token-----------" << endl;
            vector<ExprElement*>::iterator oit = opers.begin();
            while(oit != opers.end() ) {
                ExprElement* tmp = *oit;
                cout << tmp->value << " ";
                printTok(tmp->token);
                ++oit;
            }
            cout << "-----------END-----------------" << endl << endl;
        }
        // Get the next token 
        token = processToken(&expr[cur],token_str,size);
        if (debug_flag) printTok(token);
        // If the previous token was a new unassigned variable
        // this token must be an "=" (ASSIGN_TOK)
        if(must_be_asignment && (token != ASSIGN_TOK)) {
            throw ExprError("No such variable",token_str_added);
        } else { // alls well
            must_be_asignment = false;
        }
        switch((TOKEN)token) {
        case EOF_TOK:
            done = true;
            break;
        // Tokens that are currently errors 
        case ERROR_TOK:
        case STRING_CONST_TOK:
        case BUILTIN_FUNCTION_TOK :
        case CONTROL_FUNCTION_TOK :
        case OPEN_BRACKET_TOK :
        case SEMICOLON_TOK :
        case ATSIGN_TOK: // This was used in raw2a below as a variable
            throw ExprError("Syntax error (1)",&expr[cur]);
            done = true;
            break;
        // Operands ////////////////////////////////////////////////////////////////////////////////////////
        // Tokens which are constant values
        case FLOAT_TOK:
        case INTEGER_TOK:
            oper = new ExprElement;
            oper->token = token;
            oper->storage_class = CONSTANT_CLASS;
            last_oper = OPERAND;
            oper->value = stod(token_str);
            if ( oper->value == HUGE_VAL) throw ExprError("Value too large",&expr[cur]);
            post.push_back(oper);
            break;
        case CHAR_TOK :
        case CTRL_CHAR_TOK :
        case OCT_CHAR_TOK :
        case HEX_CHAR_TOK :
            oper = new ExprElement;
            oper->token = token;
            oper->storage_class = CONSTANT_CLASS;
            oper->value = (double)string2Byte(token_str);
            last_oper = OPERAND;
            post.push_back(oper);
            break;
        case OCTAL_TOK:
        case HEX_TOK:
            oper = new ExprElement;
            oper->token = token;
            oper->storage_class = CONSTANT_CLASS;
            last_oper = OPERAND;
            oper->value = stoll(token_str,NULL,0);
            if ( oper->value == HUGE_VAL) throw ExprError("Value too large",&expr[cur]);
            post.push_back(oper);
            break;
        case BIN_TOK:
            oper = new ExprElement;
            oper->token = token;
            oper->storage_class = CONSTANT_CLASS;
            last_oper = OPERAND;
            oper->value = stoll(&token_str[2],NULL,2);
            if ( oper->value == HUGE_VAL) throw ExprError("Value too large",&expr[cur]);
            post.push_back(oper);
            break;
        // Variables
        case VARIABLE_TOK:
        case SHELL_VARIABLE_TOK:
        // case ATSIGN_TOK: // This was used in raw2a
            oper = new ExprElement;
            oper->token = VARIABLE_TOK;
            oper->storage_class = VARIABLE_CLASS;
            last_oper = OPERAND;
            // Check to see if this variable has been allocated
            if ( var_list->find(token_str) == var_list->end() ) {
                if ( ! allow_allocation ) throw ExprError("Unknown variable",token_str);
                (*var_list)[token_str]=0;
                must_be_asignment = true;
                token_str_added = token_str;
            } else { // It has
                oper->value = var_list->at(token_str);
            }
            oper->token_str = token_str;
            post.push_back(oper);
            break;
        // Determine if it's add or subtract or if it's designating a values polarity 
        case UNARY_MINUS_TOK:
        case UNARY_PLUS_TOK:
            if(last_oper == OPERAND) {
                oper = new ExprElement;
                if(token == UNARY_PLUS_TOK) oper->token = ADD_TOK;
                else oper->token = SUBTRACT_TOK;
            } else {
                oper = new ExprElement;
                oper->token = token;
            }
            oper->storage_class = OPERATOR_CLASS;
            last_oper = OPERATOR;
            while(!opers.empty() && (tokenPrecedence((int)oper->token) < tokenPrecedence((int)opers.back()->token))) {
                post.push_back(opers.back());
                opers.pop_back();
            }
            opers.push_back(oper);
            break;
        // Operator tokens ///////////////////////////////////////////////////////////////////////////////
        // Increment and decrement 
        case POST_INC_TOK:
        case POST_DEC_TOK:
            if(last_oper == OPERAND) {
                oper = new ExprElement;
                oper->token = token;
            } else {
                oper = new ExprElement;
                if(token == POST_INC_TOK) oper->token = PRE_INC_TOK;
                else oper->token = PRE_DEC_TOK;
            }
            oper->storage_class = OPERATOR_CLASS;
            last_oper = OPERATOR;
            while(!opers.empty() && (tokenPrecedence((int)oper->token) < tokenPrecedence((int)opers.back()->token))) {
                post.push_back(opers.back());
                opers.pop_back();
            }
            opers.push_back(oper);
            break;
        // Parentheses
        case OPEN_PAREN_TOK :
            ++parens;
            oper = new ExprElement;
            oper->token = token;
            oper->storage_class = OPERATOR_CLASS;
            last_oper = OPERATOR;
            opers.push_back(oper);
            break;
        case CLOSE_PAREN_TOK:
            if(!closeParenFun(parens)) {
                if(--parens < 0) {
                    done = true;
                    break;
                }
                last_oper = OPERAND;
                while(!opers.empty() && (opers.back()->token != OPEN_PAREN_TOK)) {
                    post.push_back(opers.back());
                    opers.pop_back();
                }
                if(!opers.empty() && opers.back()->token == OPEN_PAREN_TOK) {
                    delete opers.back();
                    opers.pop_back();
                }
            }
            break;
        // To separate function arguments.  
        case COMMA_TOK :
            commaFun(parens); 
            break;
        // Math functions
        case FUNCTION_TOK :
            if(!parseFun(token_str,parens)) {
                // Unknown function
                done = true;
                break;
            } // Now just drop through to the operator code
        // Every thing else a math operators
        default:
            oper = new ExprElement;         
            oper->token = token;
            oper->storage_class = OPERATOR_CLASS;
            last_oper = OPERATOR;
            oper->token_str = token_str;
            while(!opers.empty() && (tokenPrecedence((int)token) < tokenPrecedence((int)opers.back()->token))) {
                post.push_back(opers.back());
                opers.pop_back();
            }
            opers.push_back(oper);
            break;
        }
        if (debug_flag) cout << "[" << &expr[cur] << "] " << '\n';
        if (expr[cur]) cur += size; // Next token
    } while(size && !done);

    // If this is not 0 we don't have a balance of open and close parentheses.
    if(parens) throw ExprError("Syntax error"," Miss placed parentheses");
    
    // Now move the what's left on the opers stack to the post stack
    while(!opers.empty()) {
        post.push_back(opers.back());
        opers.pop_back();
    }
    

    if (debug_flag) {
        cout << "--------post->token-----------" << endl;
        vector<ExprElement*>::iterator pit = post.begin();
        while(pit != post.end() ) {
            ExprElement* tmp = *pit;
            printTok(tmp->token);
            ++pit;
        }
    }
     
    if(last_oper == OPERATOR &&  post.size() < 2) throw ExprError("Syntax error"," There must be a operand");

    return evalPostfix(post);

} // End of solve()

///////////////////////////////////////////////////////////////

double AlgebraicExpr::evalPostfix(vector<ExprElement*>& post)
{
    vector<ExprElement*> stack;
    ExprElement* oper;
    TOKEN_TYPE last_oper = OPERATOR;

    if (debug_flag) cout << "-----calculate-----" << endl; 
    vector<ExprElement*>::iterator cur_post = post.begin();
    if(post.end() != cur_post) do {
        ExprElement* operand;
        long long tmp_int = 0;
        long long tmp_int1 = 0;
        oper = *cur_post;
        if(oper->storage_class == OPERATOR_CLASS) {
            if (debug_flag) printTok(oper->token);
            switch((TOKEN)oper->token) {
            case POST_INC_TOK:
                if(stack.back()->storage_class == VARIABLE_CLASS) (*var_list)[stack.back()->token_str] += 1;
                else throw ExprError("Syntax error","++ Must follow a variable");
                break;
            case POST_DEC_TOK:
                if(stack.back()->storage_class == VARIABLE_CLASS) (*var_list)[stack.back()->token_str] -= 1;
                else throw ExprError("Syntax error","-- Must follow a variable");
                break;
            case PRE_INC_TOK :
                if(stack.back()->storage_class == VARIABLE_CLASS) (*var_list)[stack.back()->token_str] += 1;
                else throw ExprError("Syntax error","++ Must precede a variable");
                stack.back()->value += 1;
                break;
            case PRE_DEC_TOK :
                if(stack.back()->storage_class == VARIABLE_CLASS) (*var_list)[stack.back()->token_str] -= 1;
                else throw ExprError("Syntax error","-- Must precede a variable");
                stack.back()->value -= 1;
                break;
            case COMPLEMENT_TOK :
                tmp_int = (long long)stack.back()->value;
                stack.back()->value = (double)(~tmp_int);
                break;
            case UNARY_MINUS_TOK:
                stack.back()->value = -(stack.back()->value);
                break;
            case UNARY_PLUS_TOK:
                stack.back()->value = +(stack.back()->value);
                break;
            case ADD_TOK :
                operand = stack.back();
                stack.pop_back();
                stack.back()->value += operand->value;
                break;
            case SUBTRACT_TOK :
                operand = stack.back();
                stack.pop_back();
                stack.back()->value -= operand->value;
                break;
            case MULTIPLY_TOK :
                operand = stack.back();
                stack.pop_back();
                stack.back()->value *= operand->value;
                break;
            case DIVIDE_TOK :
                operand = stack.back();
                stack.pop_back();
                if(operand->value == 0) throw ExprError("Syntax error"," Dividing by zero");
                stack.back()->value /= operand->value;
                break;
            case MODULO_TOK :
                operand = stack.back();
                stack.pop_back();
                if(operand->value == 0) throw ExprError("Syntax error"," Dividing by zero");
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                stack.back()->value = (double)(tmp_int1 % tmp_int);
                break;
            case SHIFT_LEFT_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                stack.back()->value = (double)(tmp_int1 << tmp_int);
                break;
            case SHIFT_RIGHT_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                stack.back()->value = (double)(tmp_int1 >> tmp_int);
                break;
            case BIT_AND_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                stack.back()->value = (double)(tmp_int & tmp_int1);
                break;
            case BIT_XOR_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                stack.back()->value = (double)(tmp_int ^ tmp_int1);
                break;
            case BIT_OR_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                stack.back()->value = (double)(tmp_int | tmp_int1);
                break;
            case AND_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value && operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case OR_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value || operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case LESS_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value < operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case LESS_EQUAL_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value <= operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case GREATER_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value > operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case GREATER_EQUAL_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value >= operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case EQUAL_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value == operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case NOT_EQUAL_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->value != operand->value) stack.back()->value = 1;
                else stack.back()->value = 0;
                break;
            case NOT_TOK :
                if(stack.back()->value) stack.back()->value = 0;
                else stack.back()->value = 1;
                break;
            case ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = operand->value;
                else throw ExprError("Syntax error"," Only variables can be asigned values (=)");
                stack.back()->value = operand->value;
                break;
            case ADD_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = operand->value + var_list->at(stack.back()->token_str);
                else throw ExprError("Syntax error"," Only variables can be asigned values (+=)");
                stack.back()->value += operand->value;
                break;
            case SUB_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = operand->value - var_list->at(stack.back()->token_str);
                else throw ExprError("Syntax error"," Only variables can be asigned values (-=)");
                stack.back()->value -= operand->value;
                break;
            case MUL_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = operand->value * var_list->at(stack.back()->token_str);
                else throw ExprError("Syntax error"," Only variables can be asigned values (*=)");
                stack.back()->value *= operand->value;
                break;
            case DIV_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                if(operand->value == 0) throw ExprError("Syntax error"," Dividing by zero");
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = operand->value / var_list->at(stack.back()->token_str);
                else throw ExprError("Syntax error"," Only variables can be asigned values (/=)");
                stack.back()->value /= operand->value;
                break;
            case MOD_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                if(operand->value == 0) throw ExprError("Syntax error"," Dividing by zero");
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = (double)(tmp_int1 % tmp_int);
                else throw ExprError("Syntax error"," Only variables ca be asigned values (%=)");
                stack.back()->value = (double)(tmp_int1 % tmp_int);
                break;
            case SH_LEFT_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = (double)(tmp_int1 << tmp_int);
                else throw ExprError("Syntax error"," Only variables ca be asigned values (<<=)");
                stack.back()->value = (double)(tmp_int1 << tmp_int);
                break;
            case SH_RIGHT_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = (double)(tmp_int1 >> tmp_int);
                else throw ExprError("Syntax error"," Only variables ca be asigned values (>>=)");
                stack.back()->value = (double)(tmp_int1 >> tmp_int);
                break;
            case AND_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = (double)(tmp_int1 & tmp_int);
                else throw ExprError("Syntax error"," Only variables ca be asigned values (&=)");
                stack.back()->value = (double)(tmp_int1 & tmp_int);
                break;
            case OR_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = (double)(tmp_int1 | tmp_int);
                else throw ExprError("Syntax error"," Only variables ca be asigned values (|=)");
                stack.back()->value = (double)(tmp_int1 | tmp_int);
                break;
            case XOR_ASSIGN_TOK :
                operand = stack.back();
                stack.pop_back();
                tmp_int = (long long)operand->value;
                tmp_int1 = (long long)stack.back()->value;
                if(stack.back()->storage_class == VARIABLE_CLASS)
                    (*var_list)[stack.back()->token_str] = (double)(tmp_int1 ^ tmp_int);
                else throw ExprError("Syntax error"," Only variables ca be asigned values (^=)");
                stack.back()->value = (double)(tmp_int1 ^ tmp_int);
                break;
            case COLON_TOK :
                ++cur_post;
                oper = *cur_post;
                if(oper->token == QUESTION_TOK) {
                    ExprElement* false_op = stack.back();
                    stack.pop_back();
                    ExprElement* true_op = stack.back();
                    stack.pop_back();
                    operand = stack.back();
                    stack.pop_back();
                    if(operand->value) {
                        stack.push_back(true_op);
                    } else {
                        stack.push_back(false_op);
                    }
                } else throw ExprError("Syntax error"," Miss placed colon");
                break;
            case QUESTION_TOK :
                throw ExprError("Syntax error"," Miss placed question mark");
                break;
            case FUNCTION_TOK :
                exeFun(oper->token_str,stack);
                break;
            default:
                throw ExprError("Syntax error"," Unknown operator");
                break;
            }
            last_oper = OPERATOR;
        } else {
            if (debug_flag) {
                cout << oper->value;
                if(oper->storage_class == VARIABLE_CLASS) cout << " [" << oper->token_str << "]" << endl;
                else cout << endl;
            }
            last_oper = OPERAND;
            stack.push_back(oper);
        }
    } while(++cur_post != post.end());
    if (debug_flag) {
        cout << "---------var_list----------" << endl;
        for (auto const& x : *var_list) {
            std::cout << x.first  // variable name
                      << ": \t" 
                      << x.second // variable value 
                      << endl ;
        }
    }
    double return_value = stack.back()->value;
    // Clean up the the postfix list
    vector<ExprElement*>::iterator pit = post.begin();
    while(pit != post.end() ) {
        ExprElement* tmp = *pit;
        delete tmp;
        ++pit;
    }
     
     
    if ( return_value == HUGE_VAL) throw ExprError("Value too large"," Final value");

    return return_value;

} // End of evalPostfix()

///////////////////////////////////////////////////////////////

TOKEN AlgebraicExpr::processToken(const string& expr,string& token_str,int& size)
{
    TOKEN token;
        
    if(expr.size() == 0) return(EOF_TOK);
    if(!(token = (TOKEN)lexer.lex(expr))) {
        throw ExprError("Syntax error (2)",expr);
    }
    token_str = lexer.text();
    if(token == SHELL_VARIABLE_TOK) {
        token_str.erase(0,1); // Remove the '$'
    }
    size = lexer.size();
    return(token);

} // End of processToken()

///////////////////////////////////////////////////////////////

char AlgebraicExpr::string2Byte(string s)
{
    string p;
    char byte = 0;

    switch(s[0]) {
    case 0 :
        byte = 0;
        break;
    case '\'' :
        if(s[1] != '\\' ) {
            byte = s[1];
        } else switch(s[2]) {
                case 'a' : byte = '\a'; break;
                case 'b' : byte = '\b'; break;
                case 'f' : byte = '\f'; break;
                case 'n' : byte = '\n'; break;
                case 'r' : byte = '\r'; break;
                case 't' : byte = '\t'; break;
                case 'v' : byte = '\v'; break;
                case '\'' : byte = '\''; break;
                case '\"' : byte = '\"'; break;
                case '\\' : byte = '\\'; break;
                    break;
                default:
                    s[1] = '0';
                    string::size_type sz = 0;
                    long n;
                    n = stol(&s[1],&sz,0);
                    if(n > 0xff) throw ExprError("Too big for a Byte",s);
                    byte = (char)n;
        }
        break;
    default:
        string::size_type sz = 0;
        long n;
        n = stol(s,&sz,0);
        if(n > 0xff) throw ExprError("Too big for a Byte",s);
        byte = (char)n;
    }
    return(byte);
    
} // End of string2Byte()

///////////////////////////////////////////////////////////////
///////////////////// END OF FILE /////////////////////////////
///////////////////////////////////////////////////////////////
