/////////////////////////////////////////////////////////////////////////////////////
///
/// algebraicexpr.h
///
/// Acme Software Works, Inc.
///
/// Created January 27, 2004 by Don Dugger
///
/// <PRE>
/// Copyright 2019 Acme Software Works, Inc.
/// 
/// Redistribution and use in source and binary forms, with or without
/// modification, are permitted provided that the following conditions are met:
/// 
/// 1. Redistributions of source code must retain the above copyright notice,
///    this list of conditions and the following disclaimer.
/// 
/// 2. Redistributions in binary form must reproduce the above copyright notice,
///    this list of conditions and the following disclaimer in the documentation
///    and/or other materials provided with the distribution.
/// 
/// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
/// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
/// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
/// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
/// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
/// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
/// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
/// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
/// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
/// OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
/// </PRE>
/////////////////////////////////////////////////////////////////////////////////////

#ifndef _algebraicexpr_h_
#define _algebraicexpr_h_

///////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <vector>
#include <map>

///////////////////////////////////////////////////////////////

#include "token_enum.h"
#include "token.h"

///////////////////////////////////////////////////////////////

using namespace std;

///////////////////////////////////////////////////////////////
/// Classes
///
/// This class evaluates mathematical expression written
/// in algebraic form. It can return the value calculated
/// or if turned on, assign the value to and allocate variable.
/// The syntax is that of standard C. 
///\n
/// The standard algebraic form also known as the infix form
/// is parsed and converted to the postfix form sometimes
/// referred to as reverse Polish notion. It is then categorized
/// and stored on an execution stack. This stack is then evaluated.
/// The primary function of this conversion form infix to postfix
/// is to reorganize the parentheses making evaluation much simpler.
///\n
/// @brief This is a algebraic expression evaluator.
////

class AlgebraicExpr {
public:
    /// The two major categories of tokens.
    enum TOKEN_TYPE { OPERATOR,OPERAND };
    /// The categories of operators, note operand is a single category.
    enum TOKEN_CATEGORY {
        NO_CAT          = 0,
        ASSIGN_CAT      = 1,
        BIT_CAT         = 2,
        COMMENT_CAT     = 3,
        COMPARE_CAT     = 4,
        FILE_CAT        = 5,
        FUNCTION_CAT    = 6,
        MATH_CAT        = 7,
        OPERAND_CAT     = 8,
        PAREN_CAT       = 9,
        PUCH_CAT        = 10,
        SPECUNARY_CAT   = 11,
        TYPE_CAT        = 12,
        UNARY_CAT       = 13
    };
    /// The storage class is used primarily for syntax checking.
    enum STORAGE_CLASS {
         VARIABLE_CLASS = 0,
         CONSTANT_CLASS = 1,
         OPERATOR_CLASS = 2
    };
private:
    /// The lexer
    /// This is the lexer produced by RE/Flex
    Lexer lexer;
    /// The size of the token.
    int size;
    /// The variable list. It is either allocated or pass in.
    map<string,double>* var_list;
    /// The flag that indicates that the builtin variable list was allocated and will need to be freed.
    bool var_list_allocated;
    /// The flag that indicates that allocating variable is to be allowed.
    /// This allow for situations that require variables to be preallocated.
    bool allow_allocation;
    /// This is a table that contains the precedence of the operators.
    /// This table is generated by the script mklexer.ksh using the file lex.tab.
    static int token_prec[];
    /// This is a table that contains the categories of the operators.
    /// This table is generated by the script mklexer.ksh using the file lex.tab.
    static TOKEN_CATEGORY token_cat[];
protected:
    /// The debug flag it currently is on off however
    /// we may want to have debug levels in the future.
    int debug_flag;
    /// This is a index to the current token the input expression.
    int cur;
    /// The expression being evaluated
    string expr;
    /// This is a class that creates an object that defines an expression element.
    /// These objects are what is stored on the execution stack.
    /// @brief  For storage of an expression element.
    class ExprElement {
    public:
        ExprElement() { 
            token_str = "";
            token = NOP_TOK;
            value = 0;
        };
        TOKEN token;
        STORAGE_CLASS storage_class;
        double value;
        string token_str;
    };
    /// Interface to the precedence table.
    int tokenPrecedence(int op) { return(token_prec[op]); };
    /// Converts single char forms ('x') to an eight bit value.
    /// It uses the library function stol() and a table of
    /// the standard escape value ie '0xff' or '\\b'. This, if 
    /// functioning correctly should duplicate the C language.
    /// @param s    A point to the string.
    /// @return The binary value.
    char string2Byte(string s);
    /// This evaluates a list of expression elements (ExprElement)
    /// in postfix order.
    /// @param post     The postfix element list.
    /// @return     The evaluated value.
    double evalPostfix(vector<ExprElement*>& post);
    /// This is the method that calls the lexer. Which determines the current token and returns it
    /// as a string and notes it's size.
    /// @param expr     The full expression.
    /// @param token    A string to store the current token.
    /// @param size     A int to store the size of the token.
    /// @return The token's enumerator.
    TOKEN processToken(const string& expr,string& token_str,int& size);
    ///////////////////////////////////////////////////////////////////////////////////////////////
    /// The virtual methods which will be defined by the math function derived class
    /// If the math function derived class is not present we throw errors. 
    ///
    virtual int parseFun(string n,int paren) {
                                throw ExprError("Syntax error (3)",&expr[cur]);
                                return(0); };
    ///
    virtual void commaFun(int paren)
                         { throw ExprError("Syntax error (4)",&expr[cur]); };
    ///
    virtual bool closeParenFun(int paren) { return(false); };
    ///
    virtual void exeFun(string n,vector<ExprElement*>& stack)
                          { throw ExprError("Syntax error (5)",&expr[cur]); };
    ///
    ///////////////////////////////////////////////////////////////////////////////////////////////
public:
    /// This constructor is used to create an AlgebraicExpr object
    /// that when used in the null form will not allow variable allocation
    /// and will allocate an internal variable list. It can optionally
    /// allow variable allocation when it is invoked with argument "true".
    /// When variable allocation is false all variable must be allocated
    /// with the method <em>AddVar()</em>.
    /// @param allow   The optional variable allocation flag.
    /// @brief  The null constructor.
    AlgebraicExpr(bool allow = false) {
                    var_list = NULL;
                    var_list_allocated = false;
                    allow_allocation = allow;
                    debug_flag = 0;
                };
    /// This constructor is used to create an AlgebraicExpr object
    /// that is given a variable list and will not allocate an internal
    /// variable list. It can optionally allow variable allocation when
    /// it is invoked with a second argument "true".
    /// When variable allocation is false variable must be allocated
    /// with the method <em>AddVar()</em> or added to the provided list.
    /// @param var     The external variable list.
    /// @param allow   The optional variable allocation flag.
    /// @brief  The constructor with the external variable list.
    AlgebraicExpr(map<string,double>& var,bool allow = true) {
                            var_list = &var;
                            var_list_allocated = false;
                            allow_allocation = allow;
                            debug_flag = 0;
                         };
    /// The destructor that must, if allocated, free the internal variable list.
    ~AlgebraicExpr() { if(var_list_allocated) delete var_list; };

    // The Public methods.
    /// This is the exception class.
    class ExprError {
    public:
        string text;
        string loc;
        ExprError(const string text,const string loc) {
            this->text = text;
            this->loc = loc;
        };
    };


    /// Change the debug flag
    void debug(int d) { debug_flag = d; };
    int  debug() { return(debug_flag); };
    /// Prints to stdout the text name of the token, used for debugging only.
    void printTok(int token);
    /// If you choose to let AlgebraicExpr allocate variables these
    /// methods are how you access them. They will also work if you
    /// supply the variable list however you would have direct access.
    /// Note that the methods are virtual so they can be overridden by
    /// derived classes.
    ///
    /// Retrieve the value of a given variable.
    /// @param n    The variable name.
    /// @return The value of the variable.
    virtual double getVar(string n) { return((*var_list)[n]); };
    /// Add a variable to the variable list. If no value is given
    /// set it to zero.
    /// @param n    The variable name.
    /// @param val  The initial value.
    /// @return If variable does not exist return true if it does false.
    virtual bool addVar(string n, double val = 0.0 ) {
                            if ( var_list->find(n) == var_list->end() ) {
                                (*var_list)[n] = val;
                                return true;
                            } else {
                                return false;
                            }
                        };
    /// Set the value of a variable to the given value.
    /// @param n    The variable name.
    /// @param val  The new value.
    /// @return If variable does exist return true if it does not false.
    virtual bool setVar(string n,double val) {
                            if ( var_list->find(n) != var_list->end() ) {
                                (*var_list)[n] = val;
                                return true;
                            } else {
                                return false;
                            }
                        };
    /// The interface to the variable allocation flag.
    /// @parma allow The allow allocation flag.
    void variableAllocation(bool allow = true) { allow_allocation = allow; };
    /// The interface to the variable allocation flag turning it on or off with the ability
    /// to set the variable list.
    /// @parma var  The vaiable list.
    void variableAllocation(map<string,double>& var, bool allow = true) {
                           if(var_list_allocated) delete var_list;
                           var_list = &var;
                           allow_allocation = allow;
                        };
    /// This is the core method that takes an algebraic expression and returns the result
    /// as well as records the values of variable that were set. There are two variations  
    /// of this method one which simply takes an expression and returns the resulting value.
    /// And one that is also given a variable list.
    /// @param ex   The expression.
    /// @return The resulting value.
    double solve(string ex);
    /// @param ex   The expression.
    /// @param var  The variable list.
    /// @return The resulting value.
    double solve(string ex,map<string,double>& var) {
                                var_list = &var;
                                return(solve(ex));
                            };
};


///////////////////////////////////////////////////////////////

#endif // _algebraicexpr_h_

